"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.MongoDbFactory = exports.MongoDb = exports.MongoCollection = void 0;
const console_1 = require("console");
const core = __importStar(require("@fluidframework/server-services-core"));
const mongodb_1 = require("mongodb");
const server_services_telemetry_1 = require("@fluidframework/server-services-telemetry");
const server_services_utils_1 = require("@fluidframework/server-services-utils");
const mongoExceptionRetryRules_1 = require("./mongoExceptionRetryRules");
const MaxFetchSize = 2000;
const MaxRetryAttempts = 3;
const InitialRetryIntervalInMs = 1000;
const errorSanitizationMessage = "REDACTED";
const errorResponseKeysAllowList = new Set([
    "_id",
    "code",
    "codeName",
    "documentId",
    "driver",
    "err",
    "errmsg",
    "errorDetails",
    "errorLabels",
    "index",
    "insertedIds",
    "message",
    "mongoTimestamp",
    "name",
    "nInserted",
    "nMatched",
    "nModified",
    "nRemoved",
    "nUpserted",
    "ok",
    "result",
    "stack",
    "tenantId",
    "type",
    "upserted",
    "writeErrors",
    "writeConcernErrors",
]);
class MongoCollection {
    constructor(collection, retryEnabled = false, telemetryEnabled = false, mongoErrorRetryAnalyzer, collectionName, apiCounterIntervalMS, apiFailureRateTerminationThreshold, apiMinimumCountToEnableTermination) {
        this.collection = collection;
        this.retryEnabled = retryEnabled;
        this.telemetryEnabled = telemetryEnabled;
        this.mongoErrorRetryAnalyzer = mongoErrorRetryAnalyzer;
        this.collectionName = collectionName;
        this.apiCounterIntervalMS = apiCounterIntervalMS;
        this.apiFailureRateTerminationThreshold = apiFailureRateTerminationThreshold;
        this.apiMinimumCountToEnableTermination = apiMinimumCountToEnableTermination;
        this.apiCounter = new server_services_utils_1.InMemoryApiCounters();
        this.failedApiCounterSuffix = ".Failed";
        setInterval(() => {
            if (!this.apiCounter.countersAreActive) {
                return;
            }
            const counters = this.apiCounter.getCounters();
            this.apiCounter.resetAllCounters();
            server_services_telemetry_1.Lumberjack.info(`MongoCollection counter for collection ${this.collectionName}`, counters);
            this.terminateBasedOnCounterThreshold(counters);
        }, this.apiCounterIntervalMS);
    }
    async aggregate(pipeline, options) {
        const req = async () => new Promise((resolve) => resolve(this.collection.aggregate(pipeline, options)));
        return this.requestWithRetry(req, "MongoCollection.aggregate");
    }
    async find(query, sort, limit = MaxFetchSize, skip) {
        const req = async () => {
            let queryCursor = this.collection.find(query).sort(sort).limit(limit);
            if (skip) {
                queryCursor = queryCursor.skip(skip);
            }
            return queryCursor.toArray();
        };
        return this.requestWithRetry(req, "MongoCollection.find", query);
    }
    async findOne(query, options) {
        const req = async () => this.collection.findOne(query, options);
        return this.requestWithRetry(req, // request
        "MongoCollection.findOne", // callerName
        query);
    }
    async findAll() {
        const req = async () => this.collection.find({}).toArray();
        return this.requestWithRetry(req, // request
        "MongoCollection.findAll");
    }
    async update(filter, set, addToSet, options = undefined) {
        const mongoOptions = { ...options, upsert: false };
        const req = async () => {
            try {
                await this.updateCore(filter, set, addToSet, mongoOptions);
            }
            catch (error) {
                this.sanitizeError(error);
                throw error;
            }
        };
        return this.requestWithRetry(req, // request
        "MongoCollection.update", // callerName
        filter);
    }
    async updateMany(filter, set, addToSet, options = undefined) {
        const mongoOptions = { ...options, upsert: false }; // This is a backward compatible change when passing in options to give more flexibility
        const req = async () => {
            try {
                await this.updateManyCore(filter, set, addToSet, mongoOptions);
            }
            catch (error) {
                this.sanitizeError(error);
                throw error;
            }
        };
        return this.requestWithRetry(req, // request
        "MongoCollection.updateMany", // callerName
        filter);
    }
    async upsert(filter, set, addToSet, options = undefined) {
        const mongoOptions = { ...options, upsert: true };
        const req = async () => {
            try {
                await this.updateCore(filter, set, addToSet, mongoOptions);
            }
            catch (error) {
                this.sanitizeError(error);
                throw error;
            }
        };
        return this.requestWithRetry(req, // request
        "MongoCollection.upsert", // callerName
        filter);
    }
    async distinct(key, query) {
        const req = async () => this.collection.distinct(key, query);
        return this.requestWithRetry(req, // request
        "MongoCollection.distinct", // callerName
        query);
    }
    async deleteOne(filter) {
        const req = async () => this.collection.deleteOne(filter);
        return this.requestWithRetry(req, // request
        "MongoCollection.deleteOne", // callerName
        filter);
    }
    async deleteMany(filter) {
        const req = async () => this.collection.deleteMany(filter);
        return this.requestWithRetry(req, // request
        "MongoCollection.deleteMany", // callerName
        filter);
    }
    async insertOne(value) {
        const req = async () => {
            try {
                const result = await this.collection.insertOne(value);
                // Older mongo driver bug, this insertedId was objectId or 3.2 but changed to any ID type consumer provided.
                // eslint-disable-next-line @typescript-eslint/no-unsafe-return
                return result.insertedId;
            }
            catch (error) {
                this.sanitizeError(error);
                throw error;
            }
        };
        return this.requestWithRetry(req, // request
        "MongoCollection.insertOne", // callerName
        value);
    }
    async insertMany(values, ordered) {
        const req = async () => {
            try {
                await this.collection.insertMany(values, {
                    ordered: false,
                });
            }
            catch (error) {
                this.sanitizeError(error);
                throw error;
            }
        };
        await this.requestWithRetry(req, // request
        "MongoCollection.insertMany");
    }
    // Create indexes with unique restriction were always failing due to :
    // 1. We have existing data
    // 2. The unique restriction not on partition key for some collections.
    // The error will continue without a DB rebuild, which is a hard and long work.
    // Also create index mostly happened at service bootstrap time. If we bubble up at that time,
    // service will failed to start. So instead we need catch the exception and log it without bubbling up.
    async createIndex(index, unique) {
        const req = async () => this.collection.createIndex(index, { unique });
        try {
            const indexName = await this.requestWithRetry(req, // request
            "MongoCollection.createIndex");
            server_services_telemetry_1.Lumberjack.info(`Created index ${indexName}`);
        }
        catch (error) {
            server_services_telemetry_1.Lumberjack.error(`Index creation failed`, undefined, error);
        }
    }
    // Create index mostly happened at service bootstrap time. If we bubble up at that time,
    // service will failed to start. So instead we need catch the exception and log it without bubbling up.
    async createTTLIndex(index, expireAfterSeconds) {
        const req = async () => this.collection.createIndex(index, { expireAfterSeconds });
        try {
            const indexName = await this.requestWithRetry(req, // request
            "MongoCollection.createTTLIndex");
            server_services_telemetry_1.Lumberjack.info(`Created TTL index ${indexName}`);
        }
        catch (error) {
            server_services_telemetry_1.Lumberjack.error(`TTL Index creation failed`, undefined, error);
        }
    }
    async findOrCreate(query, value, options = {
        returnOriginal: true,
        upsert: true,
    }) {
        const req = async () => {
            try {
                const result = await this.collection.findOneAndUpdate(query, { $setOnInsert: value }, options);
                return result.value
                    ? { value: result.value, existing: true }
                    : { value, existing: false };
            }
            catch (error) {
                this.sanitizeError(error);
                throw error;
            }
        };
        return this.requestWithRetry(req, // request
        "MongoCollection.findOrCreate", // callerName
        query);
    }
    async findAndUpdate(query, value, options = { returnDocument: "before" }) {
        const req = async () => {
            try {
                // eslint-disable-next-line @typescript-eslint/await-thenable
                const result = await this.collection.findOneAndUpdate(query, { $set: value }, options);
                return result.value
                    ? { value: result.value, existing: true }
                    : { value, existing: false };
            }
            catch (error) {
                this.sanitizeError(error);
                throw error;
            }
        };
        return this.requestWithRetry(req, // request
        "MongoCollection.findAndUpdate", // callerName
        query);
    }
    async updateCore(filter, set, addToSet, options) {
        const update = {};
        if (set) {
            update.$set = set;
        }
        if (addToSet) {
            update.$addToSet = addToSet;
        }
        return this.collection.updateOne(filter, update, options);
    }
    async updateManyCore(filter, set, addToSet, options) {
        const update = {};
        if (set) {
            update.$set = set;
        }
        if (addToSet) {
            update.$addToSet = addToSet;
        }
        return this.collection.updateMany(filter, update, options);
    }
    async requestWithRetry(request, callerName, queryOrFilter) {
        const telemetryProperties = this.getTelemetryPropertiesFromQuery(queryOrFilter);
        try {
            const result = await core.runWithRetry(request, callerName, MaxRetryAttempts, // maxRetries
            InitialRetryIntervalInMs, // retryAfterMs
            telemetryProperties, (e) => e.code === 11000, // shouldIgnoreError
            (e) => this.retryEnabled && this.mongoErrorRetryAnalyzer.shouldRetry(e), // ShouldRetry
            (error, numRetries, retryAfterInterval) => numRetries * retryAfterInterval, // calculateIntervalMs
            (error) => this.sanitizeError(error) /* onErrorFn */, this.telemetryEnabled);
            this.apiCounter.incrementCounter(callerName);
            return result;
        }
        catch (err) {
            this.apiCounter.incrementCounter(`${callerName}${this.failedApiCounterSuffix}`);
            throw err;
        }
    }
    getTelemetryPropertiesFromQuery(queryOrFilter) {
        const properties = new Map();
        if (!queryOrFilter) {
            return properties;
        }
        if (Object.prototype.hasOwnProperty.call(queryOrFilter, "_id")) {
            properties.set("id", queryOrFilter._id);
        }
        if (Object.prototype.hasOwnProperty.call(queryOrFilter, "tenantId")) {
            properties.set(server_services_telemetry_1.BaseTelemetryProperties.tenantId, queryOrFilter.tenantId);
        }
        if (Object.prototype.hasOwnProperty.call(queryOrFilter, "documentId")) {
            properties.set(server_services_telemetry_1.BaseTelemetryProperties.documentId, queryOrFilter.documentId);
        }
        return properties;
    }
    sanitizeError(error) {
        if (error) {
            try {
                Object.keys(error).forEach((key) => {
                    if (key === "_id" || /^\d+$/.test(key)) {
                        // skip mongodb's ObjectId and array indexes
                        return;
                    }
                    else if (typeof error[key] === "object") {
                        this.sanitizeError(error[key]);
                    }
                    else if (!errorResponseKeysAllowList.has(key)) {
                        error[key] = errorSanitizationMessage;
                    }
                });
            }
            catch (err) {
                server_services_telemetry_1.Lumberjack.error(`Error sanitization failed.`, undefined, err);
                throw err;
            }
        }
    }
    terminateBasedOnCounterThreshold(counters) {
        if (this.apiFailureRateTerminationThreshold > 1) {
            return; // If threshold set more than 1, meaning we should never terminate and skip followings.
        }
        let totalCount = 0;
        let totalFailedCount = 0;
        for (const [apiName, apiCounter] of Object.entries(counters)) {
            totalCount += apiCounter;
            if (apiName.endsWith(this.failedApiCounterSuffix)) {
                totalFailedCount += apiCounter;
            }
        }
        if (totalCount < this.apiMinimumCountToEnableTermination) {
            // too little requests to consider
            return;
        }
        const failureRate = totalFailedCount / totalCount;
        if (failureRate <= this.apiFailureRateTerminationThreshold) {
            // OK, no need to terminate
            return;
        }
        // This logic is to automate the process of terminates application if db become unfunctional, so
        // kubernetes would automatically handle the restart process.
        server_services_telemetry_1.Lumberjack.warning(`Failure rate ${failureRate} more than threshold ${this.apiFailureRateTerminationThreshold}`);
        process.kill(process.pid, "SIGTERM");
    }
}
exports.MongoCollection = MongoCollection;
class MongoDb {
    constructor(client, retryEnabled = false, telemetryEnabled = false, mongoErrorRetryAnalyzer, apiCounterIntervalMS, apiFailureRateTerminationThreshold, apiMinimumCountToEnableTermination) {
        this.client = client;
        this.retryEnabled = retryEnabled;
        this.telemetryEnabled = telemetryEnabled;
        this.mongoErrorRetryAnalyzer = mongoErrorRetryAnalyzer;
        this.apiCounterIntervalMS = apiCounterIntervalMS;
        this.apiFailureRateTerminationThreshold = apiFailureRateTerminationThreshold;
        this.apiMinimumCountToEnableTermination = apiMinimumCountToEnableTermination;
    }
    // eslint-disable-next-line @typescript-eslint/promise-function-async
    close() {
        return this.client.close();
    }
    on(event, listener) {
        this.client.on(event, listener);
    }
    collection(name, dbName = "admin") {
        const collection = this.client.db(dbName).collection(name);
        return new MongoCollection(collection, this.retryEnabled, this.telemetryEnabled, this.mongoErrorRetryAnalyzer, name, this.apiCounterIntervalMS, this.apiFailureRateTerminationThreshold, this.apiMinimumCountToEnableTermination);
    }
    async dropCollection(name, dbName = "admin") {
        return this.client.db(dbName).dropCollection(name);
    }
}
exports.MongoDb = MongoDb;
const DefaultMongoDbMonitoringEvents = [
    "serverOpening",
    "serverClosed",
    "serverDescriptionChanged",
    "topologyOpening",
    "topologyClosed",
    "topologyDescriptionChanged",
    "serverHeartbeatStarted",
    "serverHeartbeatSucceeded",
    "serverHeartbeatFailed",
    // "commandStarted", Comment out because this will be too often
    // "commandSucceeded", Comment out because this will be too often
    // "commandFailed", Comment out because this will be too often
    "connectionPoolCreated",
    "connectionPoolReady",
    "connectionPoolClosed",
    "connectionCreated",
    "connectionReady",
    "connectionClosed",
    // "connectionCheckOutStarted", Comment out because this will be too often
    "connectionCheckOutFailed",
    // "connectionCheckedOut", Comment out because this will be too often
    // "connectionCheckedIn", Comment out because this will be too often
    "connectionPoolCleared",
];
const DefaultHeartbeatFrequencyMS = 30000;
const DefaultKeepAliveInitialDelay = 60000;
const DefaultSocketTimeoutMS = 100000;
const DefaultConnectionTimeoutMS = 120000;
const DefaultMinHeartbeatFrequencyMS = 10000;
const DefaultApiCounterIntervalMS = 60000;
// 1 means 100%, using 2 just for safety for incorrect calculations and meaning this feature disabled
const DefaultApiFailureRateTerminationThreshold = 2;
const DefaultApiMinimumCountToEnableTermination = 30;
const DefaultServerSelectionTimeoutMS = 30000;
class MongoDbFactory {
    constructor(config) {
        this.retryEnabled = false;
        this.telemetryEnabled = false;
        this.connectionNotAvailableMode = "ruleBehavior";
        const { operationsDbEndpoint, globalDbEnabled, globalDbEndpoint, connectionPoolMinSize, connectionPoolMaxSize, directConnection, connectionNotAvailableMode, dbMonitoringEventsList, heartbeatFrequencyMS, keepAliveInitialDelay, socketTimeoutMS, connectionTimeoutMS, minHeartbeatFrequencyMS, apiCounterIntervalMS, apiFailureRateTerminationThreshold, apiMinimumCountToEnableTermination, serverSelectionTimeoutMS, } = config;
        if (globalDbEnabled) {
            this.globalDbEndpoint = globalDbEndpoint;
        }
        (0, console_1.assert)(!!operationsDbEndpoint, `No endpoint provided`);
        this.operationsDbEndpoint = operationsDbEndpoint;
        this.connectionPoolMinSize = connectionPoolMinSize;
        this.connectionPoolMaxSize = connectionPoolMaxSize;
        this.connectionNotAvailableMode = connectionNotAvailableMode ?? "ruleBehavior";
        this.directConnection = directConnection ?? false;
        this.retryEnabled = config.facadeLevelRetry || false;
        this.telemetryEnabled = config.facadeLevelTelemetry || false;
        this.retryRuleOverride = config.facadeLevelRetryRuleOverride
            ? new Map(Object.entries(config.facadeLevelRetryRuleOverride))
            : new Map();
        this.dbMonitoringEventsList = dbMonitoringEventsList ?? DefaultMongoDbMonitoringEvents;
        this.heartbeatFrequencyMS = heartbeatFrequencyMS ?? DefaultHeartbeatFrequencyMS;
        this.keepAliveInitialDelay = keepAliveInitialDelay ?? DefaultKeepAliveInitialDelay;
        this.socketTimeoutMS = socketTimeoutMS ?? DefaultSocketTimeoutMS;
        this.connectionTimeoutMS = connectionTimeoutMS ?? DefaultConnectionTimeoutMS;
        this.minHeartbeatFrequencyMS = minHeartbeatFrequencyMS ?? DefaultMinHeartbeatFrequencyMS;
        this.apiCounterIntervalMS = apiCounterIntervalMS ?? DefaultApiCounterIntervalMS;
        this.apiFailureRateTerminationThreshold =
            apiFailureRateTerminationThreshold ?? DefaultApiFailureRateTerminationThreshold;
        this.apiMinimumCountToEnableTermination =
            apiMinimumCountToEnableTermination ?? DefaultApiMinimumCountToEnableTermination;
        this.serverSelectionTimeoutMS = serverSelectionTimeoutMS ?? DefaultServerSelectionTimeoutMS;
    }
    async connect(global = false) {
        (0, console_1.assert)(!global || !!this.globalDbEndpoint, `No global endpoint provided
                 when trying to connect to global db.`);
        // Need to cast to any before MongoClientOptions due to missing properties in d.ts
        const options = {
            directConnection: this.directConnection ?? false,
            keepAlive: true,
            keepAliveInitialDelay: this.keepAliveInitialDelay,
            socketTimeoutMS: this.socketTimeoutMS,
            connectTimeoutMS: this.connectionTimeoutMS,
            heartbeatFrequencyMS: this.heartbeatFrequencyMS,
            minHeartbeatFrequencyMS: this.minHeartbeatFrequencyMS,
            serverSelectionTimeoutMS: this.serverSelectionTimeoutMS,
        };
        if (this.connectionPoolMinSize) {
            options.minPoolSize = this.connectionPoolMinSize;
        }
        if (this.connectionPoolMaxSize) {
            options.maxPoolSize = this.connectionPoolMaxSize;
        }
        const connection = await mongodb_1.MongoClient.connect(global ? this.globalDbEndpoint : this.operationsDbEndpoint, options);
        for (const monitoringEvent of this.dbMonitoringEventsList) {
            connection.on(monitoringEvent, (event) => {
                // Using an event here so that we can use geneva monitoring in the future if we want to build alerts.
                const eventWithName = { ...event, MonitoringEventName: monitoringEvent };
                const metric = server_services_telemetry_1.Lumberjack.newLumberMetric(server_services_telemetry_1.LumberEventName.MongoMonitoring, eventWithName);
                metric.success(`Event recorded for ${monitoringEvent}`);
            });
        }
        server_services_telemetry_1.Lumberjack.info("Added event listeners", this.dbMonitoringEventsList);
        const retryAnalyzer = mongoExceptionRetryRules_1.MongoErrorRetryAnalyzer.getInstance(this.retryRuleOverride, this.connectionNotAvailableMode);
        return new MongoDb(connection, this.retryEnabled, this.telemetryEnabled, retryAnalyzer, this.apiCounterIntervalMS, this.apiFailureRateTerminationThreshold, this.apiMinimumCountToEnableTermination);
    }
}
exports.MongoDbFactory = MongoDbFactory;
//# sourceMappingURL=mongodb.js.map